"use strict";

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const STATUS_SET_EXPLICITLY = Symbol('is-status-set-explicitly');
const STATUS_CODE = Symbol('status-code');
const CONTEXT = Symbol('koa-context');
const ARGUMENTED = Symbol('argumented');

const REQUEST_PROPERTIES = ctx => ({
  [ARGUMENTED]: {
    value: true
  },
  [CONTEXT]: {
    value: ctx
  }
});

const BASE_RESPONSE_PROPERTIES = {
  [ARGUMENTED]: {
    value: true
  },
  [STATUS_SET_EXPLICITLY]: {
    writable: true
  },
  statusCode: {
    get() {
      return this[STATUS_SET_EXPLICITLY] ? this[STATUS_CODE] // The default statusCode of express is `200`
      : 200;
    },

    set(code) {
      this[STATUS_SET_EXPLICITLY] = true;
      this[STATUS_CODE] = code;
    }

  }
};

const RESPONSE_PROPERTIES = ctx => {
  const res = ctx.res;
  const writeHead = res.writeHead,
        end = res.end,
        statusCode = res.statusCode;
  return _objectSpread({}, BASE_RESPONSE_PROPERTIES, {
    [CONTEXT]: {
      value: ctx
    },
    [STATUS_CODE]: {
      writable: true,
      value: statusCode
    },
    writeHead: {
      value(...args) {
        this[STATUS_SET_EXPLICITLY] = true;
        return writeHead.apply(this, args);
      }

    },
    end: {
      value(...args) {
        // Koa and Koa2 set the statusCode to `404` by default.
        // So we must do something as well as `ctx.body = body`.
        if (!this[STATUS_SET_EXPLICITLY]) {
          // set to _response.statusCode
          this.statusCode = 200;
        }

        return end.apply(this, args);
      }

    }
  });
};

const define = (host, props) => {
  // NEVER redefine properties
  if (host[ARGUMENTED]) {
    return host;
  }

  return Object.defineProperties(host, props);
};

const makeRequest = ctx => define(ctx.req, REQUEST_PROPERTIES(ctx)); // We HAVE to manipulate the vanilla OutGoing response,
// otherwise, if the response is piped, the request will stuck and hang


const makeResponse = (ctx, resolve) => define(ctx.res, RESPONSE_PROPERTIES(ctx, resolve));

const wrap = (ctx, middleware, next) => new Promise((resolve, reject) => {
  middleware(makeRequest(ctx), makeResponse(ctx), err => {
    if (err) {
      reject(err);
      return;
    }

    resolve(next());
  });
});

module.exports = middleware => (ctx, next) => wrap(ctx, middleware, next);

Object.defineProperty(module.exports, 'CONTEXT', {
  value: CONTEXT
});